/*
 *  Copyright (c) 2018 LLC LEXY. All rights reserved.
 */

#ifndef __V_NUEROCOMMUNICATOR_BASE_H_
#define __V_NUEROCOMMUNICATOR_BASE_H_

#include <chrono>
#include <memory>

#include <QObject>
#include <QDateTime>
#include <QBluetoothAddress>

#include <vnetwork/v_network_global.h>

class QTimer;
class QBluetoothAddress;
class QNetworkAccessManager;

class VAudioDevice;
class VWiFiManager;
class VAccessPoint;
class VAudioManager;
class VBrailleDisplay;
class VBluetoothManager;
class VBrailleUsbMonitor;
class VFoundBluetoothDevice;
class VNeurocommuncatorPeripheral;

class VNETWORK_LIBRARY_EXPORT VNeurocommunicatorBase : public QObject
{
    Q_OBJECT

public:
    enum
    {
        RUSSIAN_LANGUAGE = 1,
        ENGLISH_LANGUAGE,
        GERMAN_LANGUAGE
    };

    enum
    {
        NO_UPDATE_AVAILABLE = 1,
        UPDATE_AVAILABLE,
        INSTALLING_UPDATE,
        PREVIOUS_UPDATE_FAILED
    };

    VNeurocommunicatorBase( VNeurocommuncatorPeripheral* peripheral,
                            unsigned id,
                            const QString& applicationName,
                            std::size_t frameSize,
                            const QString& speakersCardName,
                            unsigned playbackSamplerate,
                            const std::chrono::milliseconds& hideTextTime,
                            bool allowBrailleInput,
                            bool disableBrailleAutoUpdate,
                            const QBluetoothAddress &scanAdapter,
                            const QBluetoothAddress &audioAdapter,
                            const QString& logsFileToSend,
                            const QString& secondAudioCardName ) noexcept;

    virtual void start( bool advertise = true ) noexcept;

protected:
    quint32 currentWiFiIpAddress( ) const noexcept;
    bool ethernetConnected( ) const noexcept;

    void setPowerLevel( quint8 level ) noexcept;
    void setCurrentIpAddress( quint32 address ) noexcept;
    void sendLocalAlarmRequest( ) noexcept;

    void registerInitialParameters( const QString& brailleDisplayName,
                                    quint64 brailleDisplayMacAddress,
                                    const QString& speakerName,
                                    quint64 speakerMacAddress,
                                    bool installUpdatesAutomatically,
                                    uint8_t updateStatus,
                                    const QString& currentVersion,
                                    const QString& updateVersion,
                                    const QString& updateDescription,
                                    bool sendingLogs,
                                    const QDateTime& lastSendLogTime ) noexcept;

    void setUpdateStatus( uint8_t status,
                          const QString& currentVersion,
                          const QString& updateVersion = QString( ),
                          const QString& updateDescription = QString( ) ) noexcept;

    virtual void processAlarmRequest( ) noexcept = 0;
    virtual void processIdentificationRequest( ) noexcept = 0;
    virtual bool processSetMinFreeSpaceRequest( quint8 minFreeSpacePercent ) noexcept = 0;
    virtual bool otherPeripheralsAreKnown( ) noexcept = 0;
    virtual void processBrailleDisplaySet( quint64 macAddress, const QString& name ) noexcept = 0;
    virtual void processNewAudioModuleAddressReceived( const QBluetoothAddress& address ) noexcept = 0;
    virtual void processBrailleDisplayRemoved( quint64 macAddress ) noexcept = 0;
    virtual void processSpeakerAdded( quint64 macAddress, const QString& name ) noexcept;
    virtual void processSpeakerRemoved( quint64 macAddress ) noexcept;
    virtual void processBrailleInput( const QString& text ) noexcept;
    virtual void processCurrentBrailleInputTextChanged( const QString& text ) noexcept;
    virtual void processSetUserLanguageRequest( quint8 language ) noexcept;
    virtual void processSendLogsRequest( bool sendLogs ) noexcept = 0;
    virtual void processSetInstallUpdatesAutomaticallyRequested( bool automatically ) noexcept = 0;
    virtual bool processInstallUpdateRequested( ) noexcept = 0;
    virtual void processLogsSent( ) noexcept = 0;

    template<typename BleAdapter>
    BleAdapter* getBleAdapter( ) noexcept {
        return static_cast<BleAdapter*>( m_bleAdapter );
    }

    virtual void processNeurocommunicatorConnected( ) noexcept;
    virtual void processNeurocommunicatorDisconnected( ) noexcept;
    virtual void prcoessBrlttyStarted( bool started ) noexcept;

    void clearTextsOnBrailleDisplay( ) noexcept;
    void removeLastPhraseOnBrailleDisplay( ) noexcept;
    void showBrailleText( const QString& text ) noexcept;

    void playSound( const int16_t* samples, size_t count ) noexcept;
    void stopSound( ) noexcept;

    void startSecondAudioStream( bool start ) noexcept;

protected slots:
    virtual void processGattError( ) noexcept = 0;

private slots:
    void onNeurocommunicatorConnected( ) noexcept;
    void onNeurocommunicatorDisconnected( ) noexcept;

    void onAlarmRequested( ) noexcept;
    void onIdentificationRequested( ) noexcept;
    void onSetMinFreeSpaceRequested( quint8 minFreeSpacePercent ) noexcept;
    void onSetUserLanguageRequested( quint8 language ) noexcept;
    void onSetSendingLogsRequested( bool sendLogs ) noexcept;

    void onActivateWiFiConnectionRequested( const QString& ssid ) noexcept;
    void onRemoveWiFiConnectionRequested( const QString& ssid ) noexcept;
    void onAddWiFiConnectionRequested( const QString& ssid,
                                       const QString& password ) noexcept;
    void onStartMonitoringWiFi( bool start ) noexcept;

    void updateWiFiAccessPoints( ) noexcept;

    void onStartMonitoringBrailleDisplays( bool start ) noexcept;
    void onAddBrailleDisplay( quint64 macAddress, bool removeOnDisconnect ) noexcept;
    void onRemoveBrailleDisplay( quint64 macAddress ) noexcept;
    void onActivateBrailleDisplay( quint64 macAddress, bool removeOnDisconnect ) noexcept;

    void onBrlttyStarted( bool started ) noexcept;

    void onBrailleDisplayDiscovered( const VFoundBluetoothDevice &display ) noexcept;
    void onSpeakerDiscovered( const VFoundBluetoothDevice &speaker ) noexcept;

    void onAudioModuleAddressGot( const QBluetoothAddress& address ) noexcept;

    void onAddSpeaker( quint64 macAddress ) noexcept;
    void onRemoveSpeaker( quint64 macAddress ) noexcept;

    void updateBluetoothSpeakers( ) noexcept;

    void onUsbBrailleDisplayConnected( const QString& driverName,
                                       const QString& devicePath ) noexcept;
    void onUsbBrailleDisplayRemoved( const QString& devicePath ) noexcept;

    void onNewBrailleInput( const QString& text ) noexcept;
    void onCurrentBrailleInputTextChanged( const QString& text ) noexcept;

    void onSetInstallUpdatesAutomaticallyRequested( bool automatically ) noexcept;
    void onInstallUpdateRequested( ) noexcept;

    void onSendLogs( ) noexcept;

private:
    Q_DISABLE_COPY( VNeurocommunicatorBase )

    enum { SEND_LOGS_TIMER_PERIOD_MS = 60000 };
    enum { SEND_LOGS_PERIOD_S = 60 * 60 };

    enum BluetoothScannerState
    {
        NOT_SCANNING,
        SCANNING_BY_REMOTE_REQUEST,
        SEARCHING_FOR_BRAILLE_DISPLAY,
        SCANNING_BY_REMOTE_REQUEST_AND_SEARCHING_FOR_BRAILLE_DISPLAY
    };

    VNeurocommuncatorPeripheral *m_bleAdapter;

    const unsigned m_id;
    QString m_logsFileToSend;

    QTimer *m_wiFiTimer;
    VWiFiManager *m_wiFiManager;
    std::shared_ptr<QList<VAccessPoint>> m_networks;

    void setDetectedBluetoothdevices( ) noexcept;
    void stopBluetoothScanByRemoteRequest( ) noexcept;
    void processBrailleDisplayActivated( ) noexcept;
    void processFoundBluetoothDevice( const VFoundBluetoothDevice &device ) noexcept;

    VBrailleDisplay *m_brailleDisplay;

    BluetoothScannerState m_bluetoothScannerState = NOT_SCANNING;
    VBluetoothManager *m_bluetoothManager;
    std::shared_ptr<QList<VFoundBluetoothDevice>> m_foundBluetoothDevices;
    quint64 m_brailleDisplayToFind = 0;
    bool m_removeBrailleDisplayOnDisconnect = false;

    std::shared_ptr<VAudioDevice> m_speakers;
    QTimer *m_bluetoothSpeakersTimer;
    std::shared_ptr<VAudioManager> m_audioManager;

    VBrailleUsbMonitor *m_brailleUsbMonitor;
    QString m_currentUsbBrailleDisplayPath;

    QString m_brailleDisplayName;
    QBluetoothAddress m_brailleDisplayMacAddress;
    QString m_speakerName;
    QBluetoothAddress m_speakerMacAddress;

    QTimer *m_sendLogsTimer;
    QNetworkAccessManager *m_networkManager;
    QDateTime m_lastSendLogTime;

    const QString m_secondAudioCardName;
};

#endif
